#include "eglWrapper_win.h"

#include <windows.h>
#include <assert.h>

#define CLASSNAME L"eglWrapper"

ATOM g_windowclass;
HWND g_hwnd;
HMODULE g_lib;
HMODULE g_gdi;
PIXELFORMATDESCRIPTOR *g_pixelformats = NULL;
EGLint g_numPixelformats = 0;
EGLint g_error = EGL_SUCCESS;

#define GLAPIENTRY _stdcall
#define GLAPIENTRYP _stdcall *

#ifndef PROC
	typedef int (WINAPI *PROC)();
#endif		

typedef PROC (GLAPIENTRYP PFNWGLGETPROCADDRESS) (LPCSTR funcName);
typedef HGLRC (GLAPIENTRYP PFNWGLCREATECONTEXT) (HDC);
typedef BOOL (GLAPIENTRYP PFNWGLSHARELISTS) ( HGLRC hglrc1, HGLRC hglrc2 );
typedef BOOL (GLAPIENTRYP PFNWGLDELETECONTEXT) ( HGLRC hglrc );
typedef BOOL (GLAPIENTRYP PFNWGLMAKECURRENT) ( HDC hdc, HGLRC hglrc );
typedef HGLRC (GLAPIENTRYP PFNWGLGETCURRENTCONTEXT) ( void );
typedef BOOL (GLAPIENTRYP PFNSWAPBUFFERS) ( HDC hdc );


PFNWGLGETPROCADDRESS winGlGetProcAddress = 0x0;
PFNWGLCREATECONTEXT winGlCreateContext = 0x0;
PFNWGLSHARELISTS winGlShareLists = 0x0;
PFNWGLDELETECONTEXT winGlDeleteContext = 0x0;
PFNWGLMAKECURRENT winGlMakeCurrent = 0x0;
PFNWGLGETCURRENTCONTEXT winGlGetCurrentContext = 0x0;
PFNSWAPBUFFERS swapBuffers = 0x0;

DECLARE_HANDLE( HPBUFFERARB );
typedef HPBUFFERARB GL_Surface;

#define WGL_DRAW_TO_PBUFFER_ARB        0x202D
#define WGL_MAX_PBUFFER_PIXELS_ARB     0x202E
#define WGL_MAX_PBUFFER_WIDTH_ARB      0x202F
#define WGL_MAX_PBUFFER_HEIGHT_ARB     0x2030
#define WGL_PBUFFER_LARGEST_ARB        0x2033
#define WGL_PBUFFER_WIDTH_ARB          0x2034
#define WGL_PBUFFER_HEIGHT_ARB         0x2035
#define WGL_PBUFFER_LOST_ARB           0x2036

#define WGL_PIXEL_TYPE_ARB             0x2013
#define WGL_COLOR_BITS_ARB             0x2014
#define WGL_RED_BITS_ARB               0x2015
#define WGL_RED_SHIFT_ARB              0x2016
#define WGL_GREEN_BITS_ARB             0x2017
#define WGL_GREEN_SHIFT_ARB            0x2018
#define WGL_BLUE_BITS_ARB              0x2019
#define WGL_BLUE_SHIFT_ARB             0x201A
#define WGL_ALPHA_BITS_ARB             0x201B
#define WGL_ALPHA_SHIFT_ARB            0x201C
#define WGL_DEPTH_BITS_ARB             0x2022
#define WGL_STENCIL_BITS_ARB           0x2023
#define WGL_BIND_TO_TEXTURE_RGB_ARB    0x2070
#define WGL_BIND_TO_TEXTURE_RGBA_ARB   0x2071
#define WGL_TEXTURE_FORMAT_ARB         0x2072
#define WGL_TEXTURE_TARGET_ARB         0x2073
#define WGL_MIPMAP_TEXTURE_ARB         0x2074
#define WGL_TEXTURE_RGB_ARB            0x2075
#define WGL_TEXTURE_RGBA_ARB           0x2076
#define WGL_NO_TEXTURE_ARB             0x2077
#define WGL_TEXTURE_CUBE_MAP_ARB       0x2078
#define WGL_TEXTURE_1D_ARB             0x2079
#define WGL_TEXTURE_2D_ARB             0x207A
#define WGL_AUX_BUFFERS_ARB            0x2024
#define WGL_DOUBLE_BUFFER_ARB          0x2011
#define	WGL_SAMPLES_ARB				   0x2042

#define WGL_FRONT_LEFT_ARB             0x2083
#define WGL_FRONT_RIGHT_ARB            0x2084
#define WGL_BACK_LEFT_ARB              0x2085
#define WGL_BACK_RIGHT_ARB             0x2086

typedef const char *( GLAPIENTRYP PFNWGLGETEXTENSIONSSTRINGARBPROC )( HDC hdc );
typedef BOOL( GLAPIENTRYP PFNWGLSWAPINTERVALFARPROC )( int );
typedef BOOL( GLAPIENTRYP PFNWGLGETPIXELFORMATATTRIBIVARBPROC )( HDC hdc, int iPixelFormat, int iLayerPlane, UINT nAttributes, const int *piAttributes, int *piValues );
typedef BOOL( GLAPIENTRYP PFNWGLGETPIXELFORMATATTRIBFVARBPROC )( HDC hdc, int iPixelFormat, int iLayerPlane, UINT nAttributes, const int *piAttributes, FLOAT *pfValues );
typedef BOOL( GLAPIENTRYP PFNWGLCHOOSEPIXELFORMATARBPROC )( HDC hdc, const int *piAttribIList, const FLOAT *pfAttribFList, UINT nMaxFormats, int *piFormats, UINT *nNumFormats );
typedef HPBUFFERARB( GLAPIENTRYP PFNWGLCREATEPBUFFERARBPROC )( HDC hDC, int iPixelFormat, int iWidth, int iHeight, const int *piAttribList );
typedef HDC( GLAPIENTRYP PFNWGLGETPBUFFERDCARBPROC )( HPBUFFERARB hPbuffer );
typedef int( GLAPIENTRYP PFNWGLRELEASEPBUFFERDCARBPROC )( HPBUFFERARB hPbuffer, HDC hDC );
typedef BOOL( GLAPIENTRYP PFNWGLDESTROYPBUFFERARBPROC )( HPBUFFERARB hPbuffer );
typedef BOOL( GLAPIENTRYP PFNWGLQUERYPBUFFERARBPROC )( HPBUFFERARB hPbuffer, int iAttribute, int *piValue );
typedef BOOL( GLAPIENTRYP PFNWGLBINDTEXIMAGEARBPROC )( HPBUFFERARB hPbuffer, int iBuffer );
typedef BOOL( GLAPIENTRYP PFNWGLRELEASETEXIMAGEARBPROC )( HPBUFFERARB hPbuffer, int iBuffer );
typedef BOOL( GLAPIENTRYP PFNWGLSETPBUFFERATTRIBARBPROC )( HPBUFFERARB hPbuffer, const int *piAttribList );

PFNWGLGETEXTENSIONSSTRINGARBPROC wglGetExtensionsStringARB = 0x0;
PFNWGLSWAPINTERVALFARPROC wglSwapIntervalEXT = 0x0;
PFNWGLGETPIXELFORMATATTRIBIVARBPROC wglGetPixelFormatAttribivARB = 0x0;
PFNWGLGETPIXELFORMATATTRIBFVARBPROC wglGetPixelFormatAttribfvARB = 0x0;
PFNWGLCHOOSEPIXELFORMATARBPROC wglChoosePixelFormatARB = 0x0;
PFNWGLCREATEPBUFFERARBPROC wglCreatePbufferARB = 0x0;
PFNWGLGETPBUFFERDCARBPROC wglGetPbufferDCARB = 0x0;
PFNWGLRELEASEPBUFFERDCARBPROC wglReleasePbufferDCARB = 0x0;
PFNWGLDESTROYPBUFFERARBPROC wglDestroyPbufferARB = 0x0;
PFNWGLQUERYPBUFFERARBPROC wglQueryPbufferARB = 0x0;
PFNWGLBINDTEXIMAGEARBPROC wglBindTexImageARB = 0x0;
PFNWGLRELEASETEXIMAGEARBPROC wglReleaseTexImageARB = 0x0;
PFNWGLSETPBUFFERATTRIBARBPROC wglSetPbufferAttribARB = 0x0;

//WINGDIAPI BOOL  WINAPI wglDeleteContext(HGLRC);
//WINGDIAPI HGLRC WINAPI wglGetCurrentContext(VOID);
//WINGDIAPI PROC  WINAPI wglGetProcAddress(LPCSTR);
//WINGDIAPI BOOL  WINAPI wglMakeCurrent(HDC, HGLRC);
//WINGDIAPI BOOL  WINAPI SwapBuffers(HDC);

void ErrorExit(LPTSTR lpszFunction) 
{ 
    // Retrieve the system error message for the last-error code

    LPVOID lpMsgBuf;    
    DWORD dw = GetLastError(); 

    FormatMessage(
        FORMAT_MESSAGE_ALLOCATE_BUFFER | 
        FORMAT_MESSAGE_FROM_SYSTEM |
        FORMAT_MESSAGE_IGNORE_INSERTS,
        NULL,
        dw,
        MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
        (LPTSTR) &lpMsgBuf,
        0, NULL );

    // Display the error message and exit the process
	OutputDebugString( (LPTSTR) lpMsgBuf );   
    LocalFree(lpMsgBuf);
    ExitProcess(dw); 
}


bool isExtensionSupported( const char* extString, const char* extensions )
{
	const char *pos = 0;
	while( ( pos = strstr( extensions, extString ) ) != 0 )
	{
		char c = *( pos + strlen( extString ) );
		if( c == ' ' || c == '\0' ) return true;
		pos += strlen( extString );
	}

	return false;
}


bool init()
{
	g_lib = LoadLibrary( TEXT("opengl32") );
	if( g_lib == NULL )
	{
		g_error = EGL_NOT_INITIALIZED;
		return false;
	}

	winGlGetProcAddress = (PFNWGLGETPROCADDRESS) GetProcAddress( g_lib, "wglGetProcAddress" );
	winGlCreateContext = (PFNWGLCREATECONTEXT) GetProcAddress( g_lib, "wglCreateContext" );
	winGlShareLists = (PFNWGLSHARELISTS ) GetProcAddress( g_lib, "wglCreateContext" );
	winGlDeleteContext = (PFNWGLDELETECONTEXT) GetProcAddress( g_lib, "wglDeleteContext" );
	winGlMakeCurrent = (PFNWGLMAKECURRENT) GetProcAddress( g_lib, "wglMakeCurrent" );
	winGlGetCurrentContext = (PFNWGLGETCURRENTCONTEXT) GetProcAddress( g_lib, "wglGetCurrentContext" );
	winGlDeleteContext = (PFNWGLDELETECONTEXT) GetProcAddress( g_lib, "wglDeleteContext" );

	g_gdi = LoadLibrary( TEXT( "GDI32" ) );
	swapBuffers = (PFNSWAPBUFFERS) GetProcAddress( g_gdi, "SwapBuffers" );

	return winGlGetProcAddress != 0;
}


#ifdef __cplusplus
extern "C" {
#endif

EGLEXP EGLDisplay EGLAPIENTRY eglGetDisplay(EGLNativeDisplayType display_id)
{
	HDC  hdc = HDC(display_id); 
	wchar_t buffer[200];
	wsprintf( buffer, L"%p\n", hdc );
	OutputDebugString( buffer );
	wsprintf( buffer, L"%p\n", display_id );
	OutputDebugString( buffer );
	return (void*) display_id;
}

EGLEXP EGLBoolean EGLAPIENTRY eglInitialize(EGLDisplay dpy, EGLint *major, EGLint *minor)
{
	if( major ) *major = 1;
	if( minor ) *minor = 1;

	if( !init() ) return EGL_FALSE;

	HDC  hdc = HDC(dpy); 
	wchar_t buffer[200];
	wsprintf( buffer, L"eglInitialize 201 %p\n", hdc );
	OutputDebugString( buffer );
	wsprintf( buffer, L"eglInitialize 203 %p\n", dpy );
	OutputDebugString( buffer );

	wglGetExtensionsStringARB =( PFNWGLGETEXTENSIONSSTRINGARBPROC ) winGlGetProcAddress( "wglGetExtensionsStringARB" );
	if( wglGetExtensionsStringARB )
	{
		const char * wglExtensions = wglGetExtensionsStringARB( hdc );

		if( isExtensionSupported( "WGL_EXT_swap_control", wglExtensions ) )
			wglSwapIntervalEXT =( PFNWGLSWAPINTERVALFARPROC )winGlGetProcAddress( "wglSwapIntervalEXT" );
		if( isExtensionSupported( "WGL_ARB_pixel_format", wglExtensions ) )
		{
			wglGetPixelFormatAttribivARB =( PFNWGLGETPIXELFORMATATTRIBIVARBPROC ) winGlGetProcAddress( "wglGetPixelFormatAttribivARB" );
			wglGetPixelFormatAttribfvARB =( PFNWGLGETPIXELFORMATATTRIBFVARBPROC ) winGlGetProcAddress( "wglGetPixelFormatAttribfvARB" );
			wglChoosePixelFormatARB =( PFNWGLCHOOSEPIXELFORMATARBPROC ) winGlGetProcAddress( "wglChoosePixelFormatARB" );
		}
		if( isExtensionSupported( "WGL_ARB_pbuffer", wglExtensions ) )
		{
			wglCreatePbufferARB	=( PFNWGLCREATEPBUFFERARBPROC ) winGlGetProcAddress( "wglCreatePbufferARB" );
			wglGetPbufferDCARB =( PFNWGLGETPBUFFERDCARBPROC ) winGlGetProcAddress( "wglGetPbufferDCARB" );
			wglReleasePbufferDCARB =( PFNWGLRELEASEPBUFFERDCARBPROC ) winGlGetProcAddress( "wglReleasePbufferDCARB" );
			wglDestroyPbufferARB =( PFNWGLDESTROYPBUFFERARBPROC ) winGlGetProcAddress( "wglDestroyPbufferARB" );
			wglQueryPbufferARB =( PFNWGLQUERYPBUFFERARBPROC ) winGlGetProcAddress( "wglQueryPbufferARB" );
		}
		if( isExtensionSupported( "WGL_ARB_render_texture", wglExtensions ) )
		{
			wglBindTexImageARB =( PFNWGLBINDTEXIMAGEARBPROC ) winGlGetProcAddress( "wglBindTexImageARB" );
			wglReleaseTexImageARB =( PFNWGLRELEASETEXIMAGEARBPROC ) winGlGetProcAddress( "wglReleaseTexImageARB" );
			wglSetPbufferAttribARB =( PFNWGLSETPBUFFERATTRIBARBPROC ) winGlGetProcAddress( "wglSetPbufferAttribARB" );
		}
		if( isExtensionSupported( "WGL_ARB_multisample", wglExtensions ) )
		{
			// TODO use wglChoosePixelFormatARB instead of ChoosePixelFormat
		}
		return EGL_TRUE;
	}
	if( dpy == 0 )
	{
		g_error = EGL_BAD_DISPLAY;
		return EGL_FALSE;
	}
	return EGL_TRUE;
}

EGLEXP EGLBoolean EGLAPIENTRY eglGetConfigs(EGLDisplay dpy, EGLConfig *configs,
			 EGLint config_size, EGLint *num_config)
{
	if( !num_config )
	{
		g_error = EGL_BAD_PARAMETER;
		return EGL_FALSE;
	}

	PIXELFORMATDESCRIPTOR pfd;
	memset( &pfd, 0, sizeof( pfd ) );
	pfd.nSize        = sizeof( pfd );

	HDC  hdc = HDC(dpy); 
	wchar_t buffer[200];
	wsprintf( buffer, L"%p\n", hdc );
	OutputDebugString( buffer );
	wsprintf( buffer, L"%p\n", dpy );
	OutputDebugString( buffer );
	int  iPixelFormat = 1; 
	unsigned int maxqual=0; int maxindex=0;
	int max_bpp, max_depth, max_dbl, max_acc;

	// obtain detailed information about  
	// the device context's first pixel format  
	*num_config = DescribePixelFormat(hdc, iPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), &pfd); 
	if( *num_config == 0 ) 
	{
		ErrorExit( TEXT("DescribePixelFormat") );
		g_error = EGL_BAD_DISPLAY;
		return EGL_FALSE;
	}

	if( !g_pixelformats || ( g_pixelformats && *num_config > g_numPixelformats ) )
	{
		delete[] g_pixelformats;
		g_pixelformats = new PIXELFORMATDESCRIPTOR[*num_config];
	}	
	g_numPixelformats = *num_config;

	int nonOpenGL = 0;
	for( iPixelFormat = 1; iPixelFormat <= *num_config; ++iPixelFormat )
	{
		PIXELFORMATDESCRIPTOR *ppfd = g_pixelformats + (iPixelFormat - 1);
		memset( ppfd, 0, sizeof( PIXELFORMATDESCRIPTOR ) );
		ppfd->nSize        = sizeof( PIXELFORMATDESCRIPTOR );
		DescribePixelFormat(hdc, iPixelFormat, sizeof(PIXELFORMATDESCRIPTOR), ppfd); 
		if( (ppfd->dwFlags & PFD_SUPPORT_OPENGL) == 0 ) // Only consider opengl configs
		{
			nonOpenGL++;
			continue;
		}

		int bpp=ppfd->cColorBits;
		int depth=ppfd->cDepthBits;
		bool pal=(ppfd->iPixelType==PFD_TYPE_COLORINDEX);
		bool mcd=((ppfd->dwFlags & PFD_GENERIC_FORMAT) && (ppfd->dwFlags & PFD_GENERIC_ACCELERATED));
		bool soft=((ppfd->dwFlags & PFD_GENERIC_FORMAT) && !(ppfd->dwFlags & PFD_GENERIC_ACCELERATED));
		bool icd=(!(ppfd->dwFlags & PFD_GENERIC_FORMAT) && !(ppfd->dwFlags & PFD_GENERIC_ACCELERATED));
		bool opengl=(ppfd->dwFlags & PFD_SUPPORT_OPENGL) != 0;
		bool stereo=(ppfd->dwFlags & PFD_STEREO) != 0;
		bool window=(ppfd->dwFlags & PFD_DRAW_TO_WINDOW) != 0;
		bool bitmap=(ppfd->dwFlags & PFD_DRAW_TO_BITMAP) != 0;
		bool dbuff=(ppfd->dwFlags & PFD_DOUBLEBUFFER) != 0;
		
		unsigned int q=0;
		if ( opengl && window) q=q+0x8000;
		if ( depth > 0 ) q=q+0x4000;
		if ( dbuff ) q=q+0x2000;
		if ( mcd || icd ) q=q+0x1000;
		if (mcd || icd) q=q+0x0040; if (icd) q=q+0x0002;
		if (bpp>=16) q=q+0x0020; if (bpp==16) q=q+0x0008;
		if (depth>=16) q=q+0x0010; if (depth==16) q=q+0x0004;
		if (!pal) q=q+0x0080;
		if (bitmap) q=q+0x0001;
        if (stereo) q=q+0x0002;
		if (q>maxqual) 
		{
			maxqual=q; 
			maxindex=iPixelFormat;
			max_bpp=bpp; 
			max_depth=depth; 
			max_dbl=dbuff?1:0; 
			max_acc=soft?0:1;
		}
	}


	if( configs && config_size > 0 )
	{
		int i = 0, j = 0;
		for(; i < config_size && i < *num_config; ++i )
		{
			PIXELFORMATDESCRIPTOR *ppfd = g_pixelformats + (iPixelFormat - 1);
			if( (ppfd->dwFlags & PFD_SUPPORT_OPENGL) == 0 ) // Only consider opengl configs
			{
				++j;
				continue;
			}
			configs[i-j] = (EGLConfig) ppfd;
		}
		if( maxindex < config_size )
			configs[maxindex] = (EGLConfig) (g_pixelformats);
		configs[0] = (EGLConfig) (g_pixelformats + maxindex);
	}

	*num_config -= nonOpenGL;

	return EGL_TRUE;
}

EGLEXP EGLBoolean EGLAPIENTRY eglGetConfigAttrib(EGLDisplay dpy, EGLConfig config,
			      EGLint attribute, EGLint *value)
{
	if( !config || !value )
	{
		g_error = EGL_BAD_PARAMETER;
		return EGL_FALSE;
	}

	PIXELFORMATDESCRIPTOR* ppfd = (PIXELFORMATDESCRIPTOR*)config;

	switch( attribute )
	{
	case EGL_BUFFER_SIZE:
		*value = ppfd->cColorBits;
		return EGL_TRUE;
	case EGL_RED_SIZE:
		*value = ppfd->cRedBits;
		return EGL_TRUE;
	case EGL_GREEN_SIZE:
		*value = ppfd->cGreenBits;
		return EGL_TRUE;
	case EGL_BLUE_SIZE:
		*value = ppfd->cBlueBits;
		return EGL_TRUE;
	case EGL_ALPHA_SIZE:
		*value = ppfd->cAlphaBits;
		return EGL_TRUE;
	case EGL_CONFIG_CAVEAT:		
		if( (ppfd->dwFlags & PFD_DOUBLEBUFFER) == 0 )
			OutputDebugString( TEXT( "Config is not double buffered!\n" ) );
		*value = (ppfd->dwFlags & (PFD_GENERIC_ACCELERATED|PFD_SUPPORT_OPENGL)) == 0 ? EGL_SLOW_CONFIG : EGL_NONE;
		return EGL_TRUE;
	case EGL_CONFIG_ID:
		*value = ((PIXELFORMATDESCRIPTOR*)config) - g_pixelformats;
		return (g_pixelformats && *value > 0) ? EGL_TRUE : EGL_FALSE;
	case EGL_DEPTH_SIZE:
		*value = ppfd->cDepthBits;
		return EGL_TRUE;
	case EGL_LEVEL:

	case EGL_MAX_PBUFFER_HEIGHT:
	case EGL_MAX_PBUFFER_WIDTH:
	case EGL_MAX_PBUFFER_PIXELS:
	case EGL_NATIVE_RENDERABLE:
		*value = (ppfd->dwFlags & PFD_DRAW_TO_WINDOW) != 0 || (ppfd->dwFlags & PFD_DRAW_TO_BITMAP) != 0;		
		return EGL_FALSE;
	case EGL_NATIVE_VISUAL_ID:
		*value = ((PIXELFORMATDESCRIPTOR*)config) - g_pixelformats;
		return (g_pixelformats && *value > 0) ? EGL_TRUE : EGL_FALSE;		
	case EGL_NATIVE_VISUAL_TYPE:
		return EGL_FALSE;
	case EGL_SAMPLE_BUFFERS:
		return EGL_FALSE;
	case EGL_SAMPLES:
		return EGL_FALSE;
	case EGL_STENCIL_SIZE:
		*value = ppfd->cStencilBits;
		return EGL_TRUE;
	case EGL_SURFACE_TYPE:
		*value = 0;
		if( (ppfd->dwFlags & PFD_DRAW_TO_WINDOW) )
			*value |= EGL_WINDOW_BIT;
		if( (ppfd->dwFlags & PFD_DRAW_TO_BITMAP) )
			*value |= EGL_PIXMAP_BIT;
		return EGL_TRUE;
	case EGL_TRANSPARENT_TYPE:
		return EGL_FALSE;
	case EGL_TRANSPARENT_RED_VALUE:
		return EGL_FALSE;
	case EGL_TRANSPARENT_GREEN_VALUE:
		return EGL_FALSE;
	case EGL_TRANSPARENT_BLUE_VALUE:
		return EGL_FALSE;
	}
	return EGL_FALSE;
}


EGLEXP EGLBoolean EGLAPIENTRY eglChooseConfig(EGLDisplay dpy, const EGLint *attrib_list, EGLConfig *configs, EGLint config_size, EGLint *num_config)
{
	HDC  hdc = HDC(dpy); 

	if( !g_pixelformats && !eglGetConfigs( dpy, 0, 0, &g_numPixelformats) )		
		return EGL_FALSE;

	if( num_config ) *num_config = 0;

	int wr = 8, wg = 8, wb = 8, wa = 8;
	int wbpp=-1; // don't care. (or a positive integer)
	int wdepth=-1; // don't care. (or a positive integer)
	int wdbl=1; // we want double-buffering. (or -1 for 'don't care', or 0 for 'none')
	int wacc=1; // we want acceleration. (or -1 or 0)
	int wstencil=0; // we don't want stencil by default
	int wsurface=EGL_WINDOW_BIT;
	int wstereo=0; // we don't want stereo by default
	int i = 0;
	while( attrib_list[i] != EGL_NONE )
	{
		switch( attrib_list[i] )
		{
		case EGL_BUFFER_SIZE:
			wbpp = attrib_list[i+1];
			break;
		case EGL_DEPTH_SIZE:
			wdepth = attrib_list[i+1];
			break;
		case EGL_SURFACE_TYPE:
			wsurface = attrib_list[i+1];            
			break;
		case EGL_STENCIL_SIZE:
			wstencil = attrib_list[i+1];
			break;
		case EGL_RED_SIZE:
			wr = attrib_list[i+1];
			break;
		case EGL_GREEN_SIZE:
			wg = attrib_list[i+1];
			break;
		case EGL_BLUE_SIZE:
			wb = attrib_list[i+1];
			break;
		case EGL_ALPHA_SIZE:
			wa = attrib_list[i+1];
			break;
		}
		i+=2;
	}

	unsigned int maxqual = 0, maxindex = 0;

	int numRetConfigs = 0;
	for( int i = 0; i < g_numPixelformats; ++i )
	{
		PIXELFORMATDESCRIPTOR *ppfd = ((PIXELFORMATDESCRIPTOR*) g_pixelformats) + i;

		int bpp=ppfd->cColorBits;
		int depth=ppfd->cDepthBits;
		bool pal=(ppfd->iPixelType==PFD_TYPE_COLORINDEX);
		bool mcd=((ppfd->dwFlags & PFD_GENERIC_FORMAT) && (ppfd->dwFlags & PFD_GENERIC_ACCELERATED));
		bool soft=((ppfd->dwFlags & PFD_GENERIC_FORMAT) && !(ppfd->dwFlags & PFD_GENERIC_ACCELERATED));
		bool icd=(!(ppfd->dwFlags & PFD_GENERIC_FORMAT) && !(ppfd->dwFlags & PFD_GENERIC_ACCELERATED));
		bool opengl=(ppfd->dwFlags & PFD_SUPPORT_OPENGL) != 0;
		bool window=(ppfd->dwFlags & PFD_DRAW_TO_WINDOW) != 0;
		bool bitmap=(ppfd->dwFlags & PFD_DRAW_TO_BITMAP) != 0;
		bool dbuff=(ppfd->dwFlags & PFD_DOUBLEBUFFER);
        bool stereo=(ppfd->dwFlags & PFD_STEREO) != 0;
		//
		unsigned int q=0;
		if (opengl && window) q=q+0x8000;
		if (wdepth==-1 || (wdepth>0 && depth>0)) q=q+0x4000;
		if (wdbl==-1 || (wdbl==0 && !dbuff) || (wdbl==1 && dbuff)) q=q+0x2000;
		if (wacc==-1 || (wacc==0 && soft) || (wacc==1 && (mcd || icd))) q=q+0x1000;
		if (mcd || icd) q=q+0x0040; if (icd) q=q+0x0002;
		if (wbpp==-1 || (wbpp==bpp)) q=q+0x0800;
		if (bpp>=16) q=q+0x0020; if (bpp==16) q=q+0x0008;
		if (wdepth==-1 || (wdepth==depth)) q=q+0x0400;
		if (depth>=16) q=q+0x0010; if (depth==16) q=q+0x0004;
		if (!pal) q=q+0x0080;
		if ( (wsurface & EGL_WINDOW_BIT) && window ) q=q+0x4000;
		if ( (wsurface & EGL_PIXMAP_BIT) && bitmap ) q=q+0x4000;
		if (wstencil==ppfd->cStencilBits) q=q+0x4000;
		else if( wstencil < ppfd->cStencilBits ) q=q+0x2000;
        if (wstereo != 0 && stereo ) 
            q = q+0x4000;
        else if( wstereo == 0 && stereo )
            q = q-0x0008;
		if (bitmap) q=q+0x0001;
		if (q>maxqual) 
		{
			maxqual=q; 
			maxindex=i;
			if( configs )
			{				
				int j = 0;
				for(; j < numRetConfigs && j < config_size - 1; ++j )
					configs[j+1] = configs[j];
				configs[0] = ppfd;
				numRetConfigs = j > 1 ? j : 1;
			}
		}
	}
	if( num_config )
		*num_config = numRetConfigs;

	return EGL_TRUE;
}

EGLEXP EGLSurface EGLAPIENTRY eglCreateWindowSurface(EGLDisplay dpy, EGLConfig config,  EGLNativeWindowType win, const EGLint *attrib_list)
{
	//if( !attrib_list ) return EGL_NO_SURFACE;

	HDC  hdc = HDC(dpy); 
	PIXELFORMATDESCRIPTOR *ppfd = (PIXELFORMATDESCRIPTOR*) config;
	
	if( !hdc )
	{
		g_error = EGL_BAD_SURFACE;
		return EGL_NO_SURFACE;
	}
	
	if( !ppfd )
	{
		g_error = EGL_BAD_CONFIG;
		return EGL_NO_SURFACE;
	}
	int pixelFormat = ChoosePixelFormat( hdc, ppfd );
	if( pixelFormat == 0 ) 
	{
		g_error = EGL_BAD_CONFIG;
		return EGL_NO_SURFACE;
	}

	if( !SetPixelFormat( hdc, pixelFormat, ppfd ) )
	{
		g_error = EGL_BAD_CONFIG;
		return EGL_NO_SURFACE;
	}
	return (EGLSurface)hdc;
}

EGLEXP EGLContext EGLAPIENTRY eglGetCurrentContext()
{
	if( winGlGetCurrentContext )
		return (EGLContext*) winGlGetCurrentContext();
	else
		return EGL_NO_CONTEXT;
}

EGLEXP EGLBoolean EGLAPIENTRY eglMakeCurrent(EGLDisplay dpy, EGLSurface draw,  EGLSurface read, EGLContext ctx)
{
	if( winGlMakeCurrent )
	{
		if( winGlMakeCurrent( HDC(dpy), HGLRC( ctx ) ) == TRUE )
			return EGL_TRUE;
		g_error = EGL_BAD_CONTEXT;
		return EGL_FALSE;
	}
	else
	{
		g_error = EGL_NOT_INITIALIZED;
		return EGL_FALSE;
	}
}

EGLEXP EGLSurface EGLAPIENTRY eglCreatePbufferSurface(EGLDisplay dpy, EGLConfig config,
				   const EGLint *attrib_list)
{
	if( !wglChoosePixelFormatARB || !wglCreatePbufferARB || !wglGetPbufferDCARB || !wglQueryPbufferARB )
	{
		// "Extension required for pbuffer unsupported"
		g_error = EGL_BAD_PARAMETER;
		return EGL_NO_SURFACE;
	}

	HDC hDC = HDC( dpy );
	if( hDC == NULL )		
	{
		g_error = EGL_BAD_DISPLAY;
		return EGL_NO_SURFACE;		
	}

	// TODO use given config
	const int iAttribList[]={
		WGL_DRAW_TO_PBUFFER_ARB, TRUE,
		//WGL_COLOR_BITS_ARB, m_colorBits==32 ? 24 : m_colorBits,
		//WGL_ALPHA_BITS_ARB, m_colorBits==32 ? 8 : 0,
		//WGL_DEPTH_BITS_ARB, m_depthBits,
		//WGL_STENCIL_BITS_ARB, m_stencilBits,
		WGL_BIND_TO_TEXTURE_RGBA_ARB, TRUE,
		WGL_DOUBLE_BUFFER_ARB, FALSE,       // We don't require double buffering for PBuffers
		0,
	};

	const float fAttribList[]={	0 };

	int pbufferFlags[]={
		WGL_TEXTURE_FORMAT_ARB, WGL_TEXTURE_RGBA_ARB,
		WGL_TEXTURE_TARGET_ARB, WGL_TEXTURE_2D_ARB,
		//request mipmap space if mipmaps are to be used
		//SGIS_generate_mipmap_supported ? WGL_MIPMAP_TEXTURE_ARB : 0,
		//SGIS_generate_mipmap_supported ? true : 0,
		0
	};

	//choose pixel format
	int pixelFormat;
	unsigned int numFormats;
	if( !wglChoosePixelFormatARB( hDC, iAttribList, fAttribList, 1, &pixelFormat, &numFormats ) )
	{			
		g_error = EGL_BAD_CONFIG;
		return EGL_NO_SURFACE;
	}

	//Create the pbuffer
	// TODO use config
	int width = 512, height = 512;
	HPBUFFERARB hPBuffer = wglCreatePbufferARB( hDC, pixelFormat, width, height, pbufferFlags );
	if( !hPBuffer )
	{
		g_error = EGL_BAD_SURFACE;
		return EGL_NO_SURFACE;
	}

	//Get the pbuffer's device context
	hDC = wglGetPbufferDCARB( hPBuffer );
	if( !hDC )
	{			
		g_error = EGL_BAD_SURFACE;
		return EGL_NO_SURFACE;
	}
	return (EGLSurface) hDC;
}


EGLEXP EGLContext EGLAPIENTRY eglCreateContext(EGLDisplay dpy, EGLConfig config, EGLContext share_context, const EGLint *attrib_list)
{
	HDC hDC = HDC( dpy );
	HGLRC hGLRC = winGlCreateContext( hDC );
	// Init OpenGL rendering context
	if( hGLRC == NULL )
	{
		g_error = EGL_BAD_DISPLAY;
		return EGL_NO_CONTEXT;
	}

	if( !winGlMakeCurrent( hDC, hGLRC ) )
	{
		g_error = EGL_BAD_CONTEXT;
		winGlDeleteContext( hGLRC );
		return EGL_NO_CONTEXT;
	}

	// share textures between contexts
	if( share_context != EGL_NO_CONTEXT )
		winGlShareLists( HGLRC(share_context), hGLRC );

	return (EGLContext) hGLRC;
}

EGLEXP EGLBoolean EGLAPIENTRY eglSwapBuffers(EGLDisplay dpy, EGLSurface surface)
{
	if( swapBuffers )
		return swapBuffers( HDC( dpy ) ) == TRUE ? EGL_TRUE : EGL_FALSE;
	else
		return EGL_FALSE;
}



EGLEXP EGLBoolean EGLAPIENTRY eglDestroyContext(EGLDisplay dpy, EGLContext ctx)
{
	if( winGlDeleteContext )
		return winGlDeleteContext( HGLRC( ctx ) ) == TRUE ? EGL_TRUE : EGL_FALSE;
	else
		return EGL_FALSE;
}

EGLEXP EGLBoolean EGLAPIENTRY eglDestroySurface(EGLDisplay dpy, EGLSurface surface)
{
	return EGL_TRUE;
}

EGLEXP EGLBoolean EGLAPIENTRY eglTerminate(EGLDisplay dpy)
{
	if( !g_lib && !g_gdi ) return EGL_FALSE;

	if( g_lib ) FreeLibrary( g_lib );
	winGlGetProcAddress = 0x0;
	winGlCreateContext = 0x0;
	winGlShareLists = 0x0;
	winGlDeleteContext = 0x0;
	winGlMakeCurrent = 0x0;
	winGlGetCurrentContext = 0x0;
	winGlDeleteContext = 0x0;
	g_lib = 0;

	if( g_gdi ) FreeLibrary( g_gdi );
	swapBuffers = 0x0;
	g_gdi = 0;

	delete[] g_pixelformats;
	g_pixelformats = 0;
	g_numPixelformats = 0;

	return EGL_TRUE;
}

EGLEXP EGLint EGLAPIENTRY eglGetError(void)
{
	if( !g_lib && !g_gdi ) return EGL_NOT_INITIALIZED;

	else
	{
		EGLint error = g_error;
		g_error = EGL_SUCCESS;
		return error;
	}
}

#ifdef TODO

//
LRESULT CALLBACK appWindowProc( HWND hwnd, UINT message, WPARAM wparam, LPARAM lparam )
{
	switch( message )
	{		
	case WM_LBUTTONUP:
		break;
	}
	return DefWindowProc( hwnd,message,wparam,lparam );
}

#ifdef UNICODE
bool createWindow( const wchar_t* title, int x, int y, int width, int height, bool fullscreen )
#else
bool createWindow( const char* title, int x, int y, int width, int height, bool fullscreen )
#endif
{
	WNDCLASS wndclass;
	// register window class( put in pointer to ourself )
	wndclass.style = CS_HREDRAW|CS_VREDRAW;
	wndclass.lpfnWndProc = &appWindowProc;
	wndclass.cbClsExtra = 4;
	wndclass.cbWndExtra = 0;
	wndclass.hInstance = 0;
	wndclass.hIcon = 0;
	wndclass.hCursor = LoadCursor( NULL,IDC_ARROW );
	wndclass.hbrBackground = NULL;
	wndclass.lpszMenuName = 0;
	wndclass.lpszClassName = CLASSNAME;
	g_windowclass = RegisterClass( &wndclass );
	if( !g_windowclass ) return false;

	g_hwnd=CreateWindowEx( 0, CLASSNAME, title, WS_POPUP|WS_MAXIMIZE, x, y, width, height, NULL, NULL, wndclass.hInstance, 0 );
	if( !g_hwnd )
	{
		UnregisterClass( CLASSNAME, wndclass.hInstance );
		return false;
	}

	if( fullscreen ) ShowWindow( g_hwnd,SW_MAXIMIZE );
	else ShowWindow( g_hwnd, SW_NORMAL );
	
	return g_hwnd != NULL;
}
#endif

#ifdef __cplusplus
}
#endif
